@@ -259,6 +259,12 @@ noregparm void *WIN_FUNC(_win_memset,3)
 	return memset(s, c, count);
 }
 
+noregparm int WIN_FUNC(_win_memcmp,3)
+	(void *s1, void *s2, SIZE_T n)
+{
+	return memcmp(s1, s2, n);
+}
+
 noregparm void WIN_FUNC(_win_srand,1)
 	(UINT seed)
 {
@@ -364,7 +370,7 @@ void dump_bytes(const char *ctx, const u8 *from, int len)
 	int i, j;
 	u8 *buf;
 
-	buf = kmalloc(len * 3 + 1, gfp_irql());
+	buf = kmalloc(len * 3 + 1, irql_gfp());
 	if (!buf) {
 		ERROR("couldn't allocate memory");
 		return;
