@@ -145,14 +145,14 @@ static int load_sys_files(struct wrap_driver *driver,
 {
 	int i, err;
 
-	TRACE1("num_pe_images = %d", load_driver->nr_sys_files);
+	TRACE1("num_pe_images = %d", load_driver->num_sys_files);
 	TRACE1("loading driver: %s", load_driver->name);
 	strncpy(driver->name, load_driver->name, sizeof(driver->name));
 	driver->name[sizeof(driver->name)-1] = 0;
 	TRACE1("driver: %s", driver->name);
 	err = 0;
 	driver->num_pe_images = 0;
-	for (i = 0; i < load_driver->nr_sys_files; i++) {
+	for (i = 0; i < load_driver->num_sys_files; i++) {
 		struct pe_image *pe_image;
 		pe_image = &driver->pe_images[driver->num_pe_images];
 
@@ -217,7 +217,7 @@ static int load_sys_files(struct wrap_driver *driver,
 		err = -EINVAL;
 	}
 
-	if (driver->num_pe_images < load_driver->nr_sys_files || err) {
+	if (driver->num_pe_images < load_driver->num_sys_files || err) {
 		for (i = 0; i < driver->num_pe_images; i++)
 			if (driver->pe_images[i].image)
 				vfree(driver->pe_images[i].image);
@@ -280,8 +280,8 @@ struct wrap_bin_file *get_bin_file(char *bin_file_name)
 		if (ret) {
 			up(&loader_mutex);
 			ERROR("couldn't load file %s/%s; check system log "
-			      "for messages from 'loadndisdriver'",
-			      driver->name, driver->bin_files[i].name);
+			      "for messages from 'loadndisdriver' (%d)",
+			      driver->name, driver->bin_files[i].name, ret);
 			EXIT1(return NULL);
 		}
 		wait_for_completion(&loader_complete);
@@ -351,22 +351,25 @@ static int load_bin_files_info(struct wrap_driver *driver,
 	struct wrap_bin_file *bin_files;
 	int i;
 
-	ENTER1("loading bin files for driver %s", load_driver->name);
-	bin_files = kmalloc(load_driver->nr_bin_files * sizeof(*bin_files),
+	ENTER1("%s, %d", load_driver->name, load_driver->num_bin_files);
+	driver->num_bin_files = 0;
+	driver->bin_files = NULL;
+	if (load_driver->num_bin_files == 0)
+		EXIT1(return 0);
+	bin_files = kzalloc(load_driver->num_bin_files * sizeof(*bin_files),
 			    GFP_KERNEL);
 	if (!bin_files) {
 		ERROR("couldn't allocate memory");
 		EXIT1(return -ENOMEM);
 	}
-	memset(bin_files, 0, load_driver->nr_bin_files * sizeof(*bin_files));
 
-	for (i = 0; i < load_driver->nr_bin_files; i++) {
+	for (i = 0; i < load_driver->num_bin_files; i++) {
 		strncpy(bin_files[i].name, load_driver->bin_files[i].name,
 			sizeof(bin_files[i].name));
 		bin_files[i].name[sizeof(bin_files[i].name)-1] = 0;
 		TRACE2("loaded bin file %s", bin_files[i].name);
 	}
-	driver->num_bin_files = load_driver->nr_bin_files;
+	driver->num_bin_files = load_driver->num_bin_files;
 	driver->bin_files = bin_files;
 	EXIT1(return 0);
 }
@@ -375,23 +378,22 @@ static int load_bin_files_info(struct wrap_driver *driver,
 static int load_settings(struct wrap_driver *wrap_driver,
 			 struct load_driver *load_driver)
 {
-	int i, nr_settings;
+	int i, num_settings;
 
 	ENTER1("%p, %p", wrap_driver, load_driver);
 
-	nr_settings = 0;
-	for (i = 0; i < load_driver->nr_settings; i++) {
+	num_settings = 0;
+	for (i = 0; i < load_driver->num_settings; i++) {
 		struct load_device_setting *load_setting =
 			&load_driver->settings[i];
 		struct wrap_device_setting *setting;
 		ULONG data1;
 
-		setting = kmalloc(sizeof(*setting), GFP_KERNEL);
+		setting = kzalloc(sizeof(*setting), GFP_KERNEL);
 		if (!setting) {
 			ERROR("couldn't allocate memory");
 			break;
 		}
-		memset(setting, 0, sizeof(*setting));
 		strncpy(setting->name, load_setting->name,
 			sizeof(setting->name));
 		setting->name[sizeof(setting->name)-1] = 0;
@@ -413,10 +415,10 @@ static int load_settings(struct wrap_driver *wrap_driver,
 			}
 		}
 		InsertTailList(&wrap_driver->settings, &setting->list);
-		nr_settings++;
+		num_settings++;
 	}
 	/* it is not a fatal error if some settings couldn't be loaded */
-	if (nr_settings > 0)
+	if (num_settings > 0)
 		EXIT1(return 0);
 	else
 		EXIT1(return -EINVAL);
@@ -561,13 +563,12 @@ static int load_user_space_driver(struct load_driver *load_driver)
 		EXIT1(return -ENOMEM);
 	}
 	TRACE1("drv_obj: %p", drv_obj);
-	drv_obj->drv_ext = kmalloc(sizeof(*(drv_obj->drv_ext)), GFP_KERNEL);
+	drv_obj->drv_ext = kzalloc(sizeof(*(drv_obj->drv_ext)), GFP_KERNEL);
 	if (!drv_obj->drv_ext) {
 		ERROR("couldn't allocate memory");
 		ObDereferenceObject(drv_obj);
 		EXIT1(return -ENOMEM);
 	}
-	memset(drv_obj->drv_ext, 0, sizeof(*(drv_obj->drv_ext)));
 	InitializeListHead(&drv_obj->drv_ext->custom_ext);
 	if (IoAllocateDriverObjectExtension(drv_obj,
 					    (void *)WRAP_DRIVER_CLIENT_ID,
@@ -789,12 +790,11 @@ static int wrapper_ioctl(struct inode *inode, struct file *file,
 		       load_device.subdevice);
 		if (load_device.vendor) {
 			struct wrap_device *wd;
-			wd = kmalloc(sizeof(*wd), GFP_KERNEL);
+			wd = kzalloc(sizeof(*wd), GFP_KERNEL);
 			if (!wd) {
 				ret = -ENOMEM;
 				break;
 			}
-			memset(wd, 0, sizeof(*wd));
 			InitializeListHead(&wd->settings);
 			wd->dev_bus = WRAP_BUS(load_device.bus);
 			wd->vendor = load_device.vendor;
