@@ -18,6 +18,8 @@
 
 #include "ntoskernel.h"
 
+//#define ALLOW_POOL_OVERFLOW 1
+
 #define NDIS_DMA_24BITS 0
 #define NDIS_DMA_32BITS 1
 #define NDIS_DMA_64BITS 2
@@ -42,6 +44,8 @@ typedef ULONG ndis_rts_threshold;
 typedef ULONG ndis_antenna;
 typedef ULONG ndis_oid;
 
+typedef UCHAR ndis_pmkid_vavlue[16];
+
 typedef uint64_t NDIS_PHY_ADDRESS;
 
 struct ndis_sg_element {
@@ -72,10 +76,11 @@ struct ndis_phy_addr_unit {
 typedef struct mdl ndis_buffer;
 
 struct ndis_buffer_pool {
-	int max_descr;
-	int num_allocated_descr;
 	ndis_buffer *free_descr;
-	NT_SPIN_LOCK lock;
+//	NT_SPIN_LOCK lock;
+	spinlock_t lock;
+	UINT max_descr;
+	UINT num_allocated_descr;
 };
 
 #define NDIS_PROTOCOL_ID_DEFAULT	0x00
@@ -179,7 +184,6 @@ struct v4_checksum {
 		};
 		ULONG value;
 	};
-
 };
 
 struct v6_checksum {
@@ -204,6 +208,18 @@ struct ndis_task_tcp_large_send {
 	BOOLEAN ip_opts;
 };
 
+struct ndis_packet;
+
+struct ndis_packet_pool {
+	struct ndis_packet *free_descr;
+//	NT_SPIN_LOCK lock;
+	spinlock_t lock;
+	UINT max_descr;
+	UINT num_allocated_descr;
+	UINT num_used_descr;
+	UINT proto_rsvd_length;
+};
+
 struct ndis_packet_stack {
 	ULONG_PTR IM_reserved[2];
 	ULONG_PTR ndis_reserved[4];
@@ -225,7 +241,7 @@ struct ndis_packet_private {
 	UINT len;
 	ndis_buffer *buffer_head;
 	ndis_buffer *buffer_tail;
-	void *pool;
+	struct ndis_packet_pool *pool;
 	UINT count;
 	ULONG flags;
 	BOOLEAN valid_counts;
@@ -292,16 +308,6 @@ struct ndis_packet_oob_data {
 	(struct ndis_packet_oob_data *)(((void *)(packet)) +		\
 					(packet)->private.oob_offset)
 
-struct ndis_packet_pool {
-	UINT max_descr;
-	UINT num_allocated_descr;
-	UINT num_used_descr;
-	struct ndis_packet *free_descr;
-	NT_SPIN_LOCK lock;
-	UINT proto_rsvd_length;
-	struct nt_list list;
-};
-
 enum ndis_device_pnp_event {
 	NdisDevicePnPEventQueryRemoved, NdisDevicePnPEventRemoved,
 	NdisDevicePnPEventSurpriseRemoved, NdisDevicePnPEventQueryStopped,
@@ -392,14 +398,14 @@ struct net_pnp_event {
 
 struct ndis_pnp_capabilities {
 	ULONG flags;
-	struct ndis_pm_wakeup_capabilities wakeup_capa;
+	struct ndis_pm_wakeup_capabilities wakeup;
 };
 
 typedef void (*ndis_isr_handler)(BOOLEAN *recognized, BOOLEAN *queue_handler,
 				 void *handle) wstdcall;
 typedef void (*ndis_interrupt_handler)(void *ctx) wstdcall;
 
-struct miniport_char {
+struct miniport {
 	/* NDIS 3.0 */
 	UCHAR major_version;
 	UCHAR minor_version;
@@ -408,13 +414,13 @@ struct miniport_char {
 	BOOLEAN (*hangcheck)(void *ctx) wstdcall;
 	void (*disable_interrupt)(void *ctx) wstdcall;
 	void (*enable_interrupt)(void *ctx) wstdcall;
-	void (*miniport_halt)(void *ctx) wstdcall;
+	void (*mp_halt)(void *ctx) wstdcall;
 	ndis_interrupt_handler handle_interrupt;
 	NDIS_STATUS (*init)(NDIS_STATUS *error_status, UINT *medium_index,
 			    enum ndis_medium medium[], UINT medium_array_size,
 			    void *handle, void *conf_handle) wstdcall;
 	ndis_isr_handler isr;
-	NDIS_STATUS (*query)(void *ctx, ndis_oid oid, void *buffer,
+	NDIS_STATUS (*queryinfo)(void *ctx, ndis_oid oid, void *buffer,
 			     ULONG buflen, ULONG *written,
 			     ULONG *needed) wstdcall;
 	void *reconfig;
@@ -425,9 +431,8 @@ struct miniport_char {
 			       ULONG buflen, ULONG *written,
 			       ULONG *needed) wstdcall;
 	NDIS_STATUS (*tx_data)(struct ndis_packet *ndis_packet,
-			       UINT *bytes_txed, void *adapter_ctx,
-			       void *rx_ctx, UINT offset,
-			       UINT bytes_to_tx) wstdcall;
+			       UINT *bytes_txed, void *mp_ctx, void *rx_ctx,
+			       UINT offset, UINT bytes_to_tx) wstdcall;
 	/* NDIS 4.0 extensions */
 	void (*return_packet)(void *ctx, void *packet) wstdcall;
 	void (*send_packets)(void *ctx, struct ndis_packet **packets,
@@ -475,7 +480,7 @@ struct ndis_rw_lock {
 	union {
 		union ndis_rw_lock_refcount ref_count[MAXIMUM_PROCESSORS];
 		/* ndiswrapper specific */
-		volatile int count;
+		int count;
 	};
 };
 
@@ -503,12 +508,11 @@ struct alloc_shared_mem {
 	BOOLEAN cached;
 };
 
-struct ndis_miniport_block;
+struct ndis_mp_block;
 
 /* this is opaque to drivers, so we can use it as we please */
 struct ndis_mp_interrupt {
 	struct kinterrupt *kinterrupt;
-	/* Taken by ISR, DisableInterrupt and SynchronizeWithInterrupt */
 	NT_SPIN_LOCK lock;
 	union {
 		void *reserved;
@@ -517,7 +521,7 @@ struct ndis_mp_interrupt {
 	ndis_isr_handler isr;
 	ndis_interrupt_handler mp_dpc;
 	struct kdpc intr_dpc;
-	struct ndis_miniport_block *nmb;
+	struct ndis_mp_block *nmb;
 	UCHAR dpc_count;
 	BOOLEAN enable;
 	struct nt_event dpc_completed_event;
@@ -532,7 +536,7 @@ struct ndis_binary_data {
 
 enum ndis_parameter_type {
 	NdisParameterInteger, NdisParameterHexInteger,
-	NdisParameterString, NdisParameterMultiString,
+	NdisParameterString, NdisParameterMultiString, NdisParameterBinary,
 };
 
 typedef struct unicode_string NDIS_STRING;
@@ -546,7 +550,7 @@ struct ndis_configuration_parameter {
 };
 
 struct wrap_ndis_driver {
-	struct miniport_char miniport;
+	struct miniport mp;
 };
 
 /* IDs used to store extensions in driver_object's custom extension */
@@ -607,8 +611,8 @@ enum ndis_media_stream_mode {
 };
 
 enum wrapper_work {
-	LINK_STATUS_CHANGED, SET_MULTICAST_LIST, COLLECT_IW_STATS,
-	HANGCHECK, SHUTDOWN
+	LINK_STATUS_OFF, LINK_STATUS_ON, SET_MULTICAST_LIST, COLLECT_IW_STATS,
+	HANGCHECK, NETIF_WAKEQ,
 };
 
 struct encr_info {
@@ -624,7 +628,7 @@ struct ndis_essid {
 	UCHAR essid[NDIS_ESSID_MAX_SIZE];
 };
 
-enum network_infrastructure {
+enum ndis_infrastructure_mode {
 	Ndis802_11IBSS, Ndis802_11Infrastructure, Ndis802_11AutoUnknown,
 	Ndis802_11InfrastructureMax
 };
@@ -678,13 +682,13 @@ struct ndis_timer {
 	struct kdpc kdpc;
 };
 
-struct ndis_miniport_timer {
+struct ndis_mp_timer {
 	struct nt_timer nt_timer;
 	struct kdpc kdpc;
 	DPC func;
 	void *ctx;
-	struct ndis_miniport_block *nmb;
-	struct ndis_miniport_timer *next;
+	struct ndis_mp_block *nmb;
+	struct ndis_mp_timer *next;
 };
 
 typedef struct cm_partial_resource_list NDIS_RESOURCE_LIST;
@@ -725,18 +729,27 @@ struct auth_encr_capa {
 	unsigned long encr;
 };
 
-enum driver_type { DRIVER_WIRELESS = 1, DRIVER_ETHERNET, };
+struct ndis_pmkid_candidate {
+	mac_address bssid;
+	DWORD flags;
+};
+
+struct ndis_pmkid_candidate_list {
+	ULONG version;
+	ULONG num_candidates;
+	struct ndis_pmkid_candidate candidates[1];
+};
 
 /*
  * This struct contains function pointers that the drivers references
  * directly via macros, so it's important that they are at the correct
  * position.
  */
-struct ndis_miniport_block {
+struct ndis_mp_block {
 	void *signature;
-	struct ndis_miniport_block *next;
+	struct ndis_mp_block *next;
 	struct driver_object *drv_obj;
-	void *adapter_ctx;
+	void *mp_ctx;
 	struct unicode_string name;
 	struct ndis_bind_paths *bindpaths;
 	void *openqueue;
@@ -748,7 +761,7 @@ struct ndis_miniport_block {
 	UCHAR assigned_cpu;
 	NT_SPIN_LOCK lock;
 	enum ndis_request_type *mediarequest;
-	struct ndis_miniport_interrupt *interrupt;
+	struct ndis_mp_interrupt *interrupt;
 	ULONG flags;
 	ULONG pnp_flags;
 	struct nt_list packet_list;
@@ -756,7 +769,7 @@ struct ndis_miniport_block {
 	struct ndis_packet *return_packet_queue;
 	ULONG request_buffer;
 	void *set_mcast_buffer;
-	struct ndis_miniport_block *primary_miniport;
+	struct ndis_mp_block *primary_mp;
 	void *wrapper_ctx;
 	void *bus_data_ctx;
 	ULONG pnp_capa;
@@ -784,7 +797,7 @@ struct ndis_miniport_block {
 	struct device_object *next_device;
 	void *mapreg;
 	void *call_mgraflist;
-	void *miniport_thread;
+	void *mp_thread;
 	void *setinfobuf;
 	USHORT setinfo_buf_len;
 	USHORT max_send_pkts;
@@ -815,8 +828,8 @@ struct ndis_miniport_block {
 	void *status_complete;
 	void *td_complete;
 
-	void *query_complete;
-	void *set_complete;
+	void *queryinfo_complete;
+	void *setinfo_complete;
 	void *wan_tx_complete;
 	void *wan_rx;
 	void *wan_rx_complete;
@@ -825,7 +838,7 @@ struct ndis_miniport_block {
 };
 
 struct wrap_ndis_device {
-	struct ndis_miniport_block *nmb;
+	struct ndis_mp_block *nmb;
 	struct wrap_device *wd;
 	struct net_device *net_dev;
 	void *shutdown_ctx;
@@ -844,14 +857,14 @@ struct wrap_ndis_device {
 	u8 tx_ring_start;
 	u8 tx_ring_end;
 	u8 is_tx_ring_full;
-	NT_SPIN_LOCK tx_ring_lock;
+	u8 tx_ok;
+	spinlock_t tx_ring_lock;
 	struct semaphore tx_ring_mutex;
 	unsigned int max_tx_packets;
-	u8 tx_ok;
-	struct semaphore ndis_comm_mutex;
-	struct task_struct *ndis_comm_task;
-	s8 ndis_comm_done;
-	NDIS_STATUS ndis_comm_status;
+	struct semaphore ndis_req_mutex;
+	struct task_struct *ndis_req_task;
+	int ndis_req_done;
+	NDIS_STATUS ndis_req_status;
 	ULONG packet_filter;
 
 	ULONG sg_dma_size;
@@ -867,46 +880,35 @@ struct wrap_ndis_device {
 	char nick[IW_ESSID_MAX_SIZE];
 	struct ndis_essid essid;
 	struct auth_encr_capa capa;
-	enum authentication_mode auth_mode;
-	enum encryption_status encr_mode;
-	enum network_infrastructure infrastructure_mode;
+	enum ndis_infrastructure_mode infrastructure_mode;
+	int max_pmkids;
 	int num_pmkids;
+	struct ndis_pmkid *pmkids;
 	mac_address mac;
 	struct proc_dir_entry *procfs_iface;
 
 	work_struct_t wrap_ndis_work;
 	unsigned long wrap_ndis_pending_work;
 	UINT attributes;
-	int iw_auth_set;
 	int iw_auth_wpa_version;
 	int iw_auth_cipher_pairwise;
 	int iw_auth_cipher_group;
 	int iw_auth_key_mgmt;
-	int iw_auth_80211_auth_alg;
+	int iw_auth_80211_alg;
 	struct ndis_packet_pool *tx_packet_pool;
 	struct ndis_buffer_pool *tx_buffer_pool;
 	int multicast_size;
 	struct v4_checksum rx_csum;
 	struct v4_checksum tx_csum;
 	enum ndis_physical_medium physical_medium;
-	u32 ndis_wolopts;
-	struct nt_list wrap_timer_list;
-	char netdev_name[IFNAMSIZ];
+	ULONG ndis_wolopts;
+	struct nt_slist wrap_timer_slist;
 	int drv_ndis_version;
+	struct ndis_pnp_capabilities pnp_capa;
+	char netdev_name[IFNAMSIZ];
 };
 
-struct ndis_pmkid_candidate {
-	mac_address bssid;
-	unsigned long flags;
-};
-
-struct ndis_pmkid_candidate_list {
-	unsigned long version;
-	unsigned long num_candidates;
-	struct ndis_pmkid_candidate candidates[1];
-};
-
-irqreturn_t ndis_isr(int irq, void *data ISR_PT_REGS_PARAM_DECL);
+BOOLEAN ndis_isr(struct kinterrupt *kinterrupt, void *ctx) wstdcall;
 
 int ndis_init(void);
 void ndis_exit(void);
@@ -932,39 +934,39 @@ void NdisAllocateBuffer(NDIS_STATUS *status, ndis_buffer **buffer,
 			struct ndis_buffer_pool *pool, void *virt,
 			UINT length) wstdcall;
 void NdisFreeBuffer(ndis_buffer *descr) wstdcall;
-void NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
+void NdisMIndicateReceivePacket(struct ndis_mp_block *nmb,
 				struct ndis_packet **packets,
 				UINT nr_packets) wstdcall;
-void NdisMSendComplete(struct ndis_miniport_block *nmb,
-		       struct ndis_packet *packet, NDIS_STATUS status) wstdcall;
-void NdisMSendResourcesAvailable(struct ndis_miniport_block *nmb) wstdcall;
-void NdisMIndicateStatus(struct ndis_miniport_block *nmb,
+void NdisMSendComplete(struct ndis_mp_block *nmb, struct ndis_packet *packet,
+		       NDIS_STATUS status) wstdcall;
+void NdisMSendResourcesAvailable(struct ndis_mp_block *nmb) wstdcall;
+void NdisMIndicateStatus(struct ndis_mp_block *nmb,
 			 NDIS_STATUS status, void *buf, UINT len) wstdcall;
-void NdisMIndicateStatusComplete(struct ndis_miniport_block *nmb) wstdcall;
-void NdisMQueryInformationComplete(struct ndis_miniport_block *nmb,
+void NdisMIndicateStatusComplete(struct ndis_mp_block *nmb) wstdcall;
+void NdisMQueryInformationComplete(struct ndis_mp_block *nmb,
 				   NDIS_STATUS status) wstdcall;
-void NdisMSetInformationComplete(struct ndis_miniport_block *nmb,
+void NdisMSetInformationComplete(struct ndis_mp_block *nmb,
 				 NDIS_STATUS status) wstdcall;
-void NdisMResetComplete(struct ndis_miniport_block *nmb,
-			NDIS_STATUS status, BOOLEAN address_reset) wstdcall;
+void NdisMResetComplete(struct ndis_mp_block *nmb, NDIS_STATUS status,
+			BOOLEAN address_reset) wstdcall;
 ULONG NDIS_BUFFER_TO_SPAN_PAGES(ndis_buffer *buffer) wstdcall;
 BOOLEAN NdisWaitEvent(struct ndis_event *event, UINT timeout) wstdcall;
 void NdisSetEvent(struct ndis_event *event) wstdcall;
 void NdisMDeregisterInterrupt(struct ndis_mp_interrupt *mp_interrupt) wstdcall;
-void EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
+void EthRxIndicateHandler(struct ndis_mp_block *nmb, void *rx_ctx,
 			  char *header1, char *header, UINT header_size,
 			  void *look_ahead, UINT look_ahead_size,
 			  UINT packet_size) wstdcall;
-void EthRxComplete(struct ndis_miniport_block *nmb) wstdcall;
-void NdisMTransferDataComplete(struct ndis_miniport_block *nmb,
+void EthRxComplete(struct ndis_mp_block *nmb) wstdcall;
+void NdisMTransferDataComplete(struct ndis_mp_block *nmb,
 			       struct ndis_packet *packet, NDIS_STATUS status,
 			       UINT bytes_txed) wstdcall;
-void NdisWriteConfiguration(NDIS_STATUS *status, struct ndis_miniport_block *nmb,
+void NdisWriteConfiguration(NDIS_STATUS *status, struct ndis_mp_block *nmb,
 			    struct unicode_string *key,
 			    struct ndis_configuration_parameter *param) wstdcall;
 void NdisReadConfiguration(NDIS_STATUS *status,
 			   struct ndis_configuration_parameter **param,
-			   struct ndis_miniport_block *nmb,
+			   struct ndis_mp_block *nmb,
 			   struct unicode_string *key,
 			   enum ndis_parameter_type type) wstdcall;
 
