@@ -45,10 +45,14 @@
 #include <linux/rtnetlink.h>
 #include <linux/highmem.h>
 
-#include "winnt_types.h"
+#if !defined(CONFIG_X86) && !defined(CONFIG_X86_64)
+#error "this module is for x86 or x86_64 architectures only"
+#endif
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,7)
 #include <linux/kthread.h>
+#else
+#include <linux/config.h>
 #endif
 /* Interrupt backwards compatibility stuff */
 #include <linux/interrupt.h>
@@ -79,6 +83,7 @@
 	dma_map_sg(&pci_dev->dev, sglist, nents, direction)
 #define UNMAP_SG(pci_dev, sglist, nents, direction)		\
 	dma_unmap_sg(&pci_dev->dev, sglist, nents, direction)
+#define PCI_DMA_MAP_ERROR(dma_addr) dma_mapping_error(dma_addr)
 
 #else // linux version <= 2.5.41
 
@@ -94,6 +99,7 @@
 	pci_map_sg(dev, sglist, nents, direction)
 #define UNMAP_SG(dev, sglist, nents, direction)		\
 	pci_unmap_sg(dev, sglist, nents, direction)
+#define PCI_DMA_MAP_ERROR(dma_addr) pci_dma_mapping_error(dma_addr)
 
 #include <linux/smp_lock.h>
 
@@ -109,6 +115,21 @@
 
 #endif // LINUX_VERSION_CODE
 
+#if defined(CONFIG_NET_RADIO) && !defined(CONFIG_WIRELESS_EXT)
+#define CONFIG_WIRELESS_EXT
+#endif
+
+#ifndef CONFIG_WIRELESS_EXT
+#warning "wirelss devices are not supported by this kernel"
+#endif
+
+#define prepare_wait_condition(task, var, value)	\
+do {							\
+	var = value;					\
+	task = current;					\
+	barrier();					\
+} while (0)
+
 /* Wait in wait_state (e.g., TASK_INTERRUPTIBLE) for condition to
  * become true; timeout is either jiffies (> 0) to wait or 0 to wait
  * forever.
@@ -144,37 +165,54 @@
 	ret;							\
 })
 
-#ifdef USE_OWN_WQ
+#ifdef WRAP_WQ
 
-typedef struct {
+struct workqueue_struct;
+
+struct workqueue_thread {
 	spinlock_t lock;
 	struct task_struct *task;
 	struct completion *completion;
 	char name[16];
 	int pid;
 	/* whether any work_structs pending? <0 implies quit */
-	int pending;
+	s8 pending;
 	/* list of work_structs pending */
 	struct list_head work_list;
+};
+
+typedef struct workqueue_struct {
+	u8 singlethread;
+	u8 qon;
+	int num_cpus;
+	struct workqueue_thread threads[0];
 } workqueue_struct_t;
 
 typedef struct {
 	struct list_head list;
 	void (*func)(void *data);
 	void *data;
-	/* whether/on which workqueue scheduled */
-	workqueue_struct_t *workq;
+	/* whether/on which thread scheduled */
+	struct workqueue_thread *thread;
 } work_struct_t;
 
 #define initialize_work(work, pfunc, pdata)			\
 	do {							\
 		(work)->func = (pfunc);				\
 		(work)->data = (pdata);				\
-		(work)->workq = NULL;				\
+		(work)->thread = NULL;				\
 	} while (0)
 
 #undef create_singlethread_workqueue
-#define create_singlethread_workqueue wrap_create_wq
+#define create_singlethread_workqueue(name) wrap_create_wq(name, 1, 0)
+#undef create_workqueue
+#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
+#define create_workqueue(name) wrap_create_wq(name, 0, 0)
+#else
+#define create_workqueue(name) wrap_create_wq(name, 1, 0)
+#define for_each_online_cpu(cpu) while ((cpu = 0) || 1)
+#define kthread_bind(thread, cpu) do { } while (0)
+#endif
 #undef destroy_workqueue
 #define destroy_workqueue wrap_destroy_wq
 #undef queue_work
@@ -182,20 +220,24 @@ typedef struct {
 #undef flush_workqueue
 #define flush_workqueue wrap_flush_wq
 
-workqueue_struct_t *wrap_create_wq(const char *name);
+workqueue_struct_t *wrap_create_wq(const char *name, u8 singlethread, u8 freeze);
+void wrap_destroy_wq_on(workqueue_struct_t *workq, int cpu);
 void wrap_destroy_wq(workqueue_struct_t *workq);
-void wrap_queue_work(workqueue_struct_t *workq, work_struct_t *work) wfastcall;
+int wrap_queue_work_on(workqueue_struct_t *workq, work_struct_t *work,
+		       int cpu) wfastcall;
+int wrap_queue_work(workqueue_struct_t *workq, work_struct_t *work) wfastcall;
 void wrap_cancel_work(work_struct_t *work);
+void wrap_flush_wq_on(workqueue_struct_t *workq, int cpu);
 void wrap_flush_wq(workqueue_struct_t *workq);
 typedef void *worker_param_t;
 #define worker_param_data(param, type, member) param
 
-#else // USE_OWN_WQ
+#else // WRAP_WQ
 
 typedef struct workqueue_struct workqueue_struct_t;
 typedef struct work_struct work_struct_t;
 
-#ifdef INIT_WORK_NAR
+#if defined(INIT_WORK_NAR) || defined(INIT_DELAYED_WORK_DEFERRABLE)
 #define initialize_work(work, func, data) INIT_WORK(work, func)
 typedef struct work_struct *worker_param_t;
 #define worker_param_data(param, type, member)	\
@@ -206,11 +248,11 @@ typedef void *worker_param_t;
 #define worker_param_data(param, type, member) param
 #endif // INIT_WORK_NAR
 
-#endif // USE_OWN_WQ
+#endif // WRAP_WQ
 
 struct nt_thread *wrap_worker_init(workqueue_struct_t *wq);
 
-#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
+#ifdef module_param
 #define WRAP_MODULE_PARM_INT(name, perm) module_param(name, int, perm)
 #define WRAP_MODULE_PARM_STRING(name, perm) module_param(name, charp, perm)
 #else
@@ -242,59 +284,8 @@ struct nt_thread *wrap_worker_init(workqueue_struct_t *wq);
 #define CHECKSUM_HW CHECKSUM_PARTIAL
 #endif
 
-#define atomic_unary_op(var, size, oper)				\
-do {									\
-	if (size == 1)							\
-		__asm__ __volatile__(					\
-			LOCK_PREFIX oper "b %b0\n\t" : "+m" (var));	\
-	else if (size == 2)						\
-		__asm__ __volatile__(					\
-			LOCK_PREFIX oper "w %w0\n\t" : "+m" (var));	\
-	else if (size == 4)						\
-		__asm__ __volatile__(					\
-			LOCK_PREFIX oper "l %0\n\t" : "+m" (var));	\
-	else if (size == 8)						\
-		__asm__ __volatile__(					\
-			LOCK_PREFIX oper "q %q0\n\t" : "+m" (var));	\
-	else {								\
-		extern void _invalid_op_size_(void);			\
-		_invalid_op_size_();					\
-	}								\
-} while (0)
-
-#define atomic_inc_var_size(var, size) atomic_unary_op(var, size, "inc")
-
-#define atomic_inc_var(var) atomic_inc_var_size(var, sizeof(var))
-
-#define atomic_dec_var_size(var, size) atomic_unary_op(var, size, "dec")
-
-#define atomic_dec_var(var) atomic_dec_var_size(var, sizeof(var))
-
-#define pre_atomic_add(var, i)					\
-({								\
-	typeof(var) pre;					\
-	__asm__ __volatile__(					\
-		LOCK_PREFIX "xadd %0, %1\n\t"			\
-		: "=r"(pre), "+m"(var)				\
-		: "0"(i));					\
-	pre;							\
-})
-
-#define post_atomic_add(var, i) (pre_atomic_add(var, i) + i)
-
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 
-#if !defined(CONFIG_PREEMPT) && !defined(preempt_enable)
-extern volatile int preempt_count;
-#define preempt_enable()  do { atomic_inc_var(preempt_count); } while (0)
-#define preempt_disable()  do { atomic_dec_var(preempt_count); } while (0)
-#define in_atomic() preempt_count
-#endif
-
-#ifndef preempt_enable_no_resched
-#define preempt_enable_no_resched() preempt_enable()
-#endif
-
 #ifndef container_of
 #define container_of(ptr, type, member)					\
 ({									\
@@ -339,9 +330,16 @@ typedef u32 pm_message_t;
 
 #ifndef PCI_D0
 #define PCI_D0 0
+#endif
+
+#ifndef PCI_D3hot
 #define PCI_D3hot 3
 #endif
 
+#ifndef PCI_D3cold
+#define PCI_D3cold 3
+#endif
+
 #ifndef PM_EVENT_SUSPEND
 #define PM_EVENT_SUSPEND 2
 #endif
@@ -385,25 +383,42 @@ typedef u32 pm_message_t;
 #define set_thread_priority(thread, prio) set_user_nice(thread, prio)
 #endif
 
+#ifndef DMA_24BIT_MASK
+#define DMA_24BIT_MASK 0x0000000000ffffffULL
+#endif
+
+#ifndef DMA_30BIT_MASK
+#define DMA_30BIT_MASK 0x000000003fffffffULL
+#endif
+
+#ifndef DMA_31BIT_MASK
+#define DMA_31BIT_MASK 0x000000007fffffffULL
+#endif
+
+#ifndef DMA_32BIT_MASK
+#define DMA_32BIT_MASK 0x00000000ffffffffULL
+#endif
+
+#ifndef __GFP_DMA32
+#define __GFP_DMA32 GFP_DMA
+#endif
+
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,22)
+#define wrap_kmem_cache_create(name, size, align, flags)	\
+	kmem_cache_create(name, size, align, flags, NULL, NULL)
+#else
+#define wrap_kmem_cache_create(name, size, align, flags)	\
+	kmem_cache_create(name, size, align, flags, NULL)
+#endif
+
+#include "winnt_types.h"
 #include "ndiswrapper.h"
 #include "pe_linker.h"
 #include "wrapmem.h"
 #include "lin2win.h"
 #include "loader.h"
 
-#ifdef CONFIG_X86_64
-#define get_sp(sp) __asm__ __volatile__("mov %%rsp, %0\n\t" : "=m"(sp))
-#else
-#define get_sp(sp) __asm__ __volatile__("mov %%esp, %0\n\t" : "=m"(sp))
-#endif
-
-#define print_sp() do {				\
-		void *sp;			\
-		get_sp(sp);			\
-		TRACE1("sp: %p", sp);		\
-	} while (0)
-
-//#define DEBUG_IRQL 1
+#include "compat.h"
 
 #if !defined(CONFIG_USB) && defined(CONFIG_USB_MODULE)
 #define CONFIG_USB 1
@@ -420,6 +435,8 @@ typedef u32 pm_message_t;
 #define SECSPERDAY		86400
 #define TICKSPERJIFFY		((TICKSPERSEC + HZ - 1) / HZ)
 
+#define int_div_round(x, y) (((x) + (y - 1)) / (y))
+
 /* 1601 to 1970 is 369 years plus 89 leap days */
 #define SECS_1601_TO_1970	((369 * 365 + 89) * (u64)SECSPERDAY)
 #define TICKS_1601_TO_1970	(SECS_1601_TO_1970 * TICKSPERSEC)
@@ -427,11 +444,12 @@ typedef u32 pm_message_t;
 /* 100ns units to HZ; if sys_time is negative, relative to current
  * clock, otherwise from year 1601 */
 #define SYSTEM_TIME_TO_HZ(sys_time)					\
-	((((sys_time) <= 0) ? (((u64)HZ * (-(sys_time))) / TICKSPERSEC) : \
-	  (((s64)HZ * ((sys_time) - ticks_1601())) / TICKSPERSEC)))
+	(((sys_time) <= 0) ? \
+	 int_div_round(((u64)HZ * (-(sys_time))), TICKSPERSEC) :	\
+	 int_div_round(((s64)HZ * ((sys_time) - ticks_1601())), TICKSPERSEC))
 
-#define MSEC_TO_HZ(ms) ((ms) * HZ / 1000)
-#define USEC_TO_HZ(us) ((us) * HZ / 1000000)
+#define MSEC_TO_HZ(ms) int_div_round((ms * HZ), 1000)
+#define USEC_TO_HZ(us) int_div_round((us * HZ), 1000000)
 
 extern u64 wrap_ticks_to_boot;
 
@@ -454,7 +472,7 @@ struct wrap_export {
 #define WIN_WIN_SYMBOL(name, argc)					\
 	{#name, (generic_func) win2lin__win_ ## name ## _ ## argc}
 #define WIN_FUNC_DECL(name, argc)			\
-	typeof(name) win2lin_ ## name ## _ ## argc ;
+	typeof(name) win2lin_ ## name ## _ ## argc;
 #define WIN_FUNC_PTR(name, argc) win2lin_ ## name ## _ ## argc
 
 #else
@@ -484,10 +502,10 @@ struct pe_image {
 	IMAGE_OPTIONAL_HEADER *opt_hdr;
 };
 
-struct ndis_miniport_block;
+struct ndis_mp_block;
 
 struct wrap_timer {
-	struct nt_list list;
+	struct nt_slist slist;
 	struct timer_list timer;
 	struct nt_timer *nt_timer;
 	long repeat;
@@ -558,6 +576,7 @@ struct wrap_device {
 	union {
 		struct {
 			struct pci_dev *pdev;
+			enum device_power_state wake_state;
 #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
 			u32 pci_state[16];
 #endif
@@ -589,231 +608,198 @@ struct wrap_device {
 	(WRAP_DEVICE(dev_bus) == WRAP_BLUETOOTH_DEVICE1 ||	\
 	 WRAP_DEVICE(dev_bus) == WRAP_BLUETOOTH_DEVICE2)
 
-extern workqueue_struct_t *ndis_wq;
-#define schedule_ndis_work(work_struct) queue_work(ndis_wq, (work_struct))
+#ifdef WRAP_WQ
+#define NTOS_WQ 1
+#endif
 
-extern workqueue_struct_t *wrapndis_wq;
-#define schedule_wrapndis_work(work_struct)	\
-	queue_work(wrapndis_wq, (work_struct))
+#define NTOS_WQ 1
 
-#define USE_NTOS_WQ 1
-
-#ifdef USE_NTOS_WQ
+#ifdef NTOS_WQ
 extern workqueue_struct_t *ntos_wq;
-#define schedule_ntos_work(work_struct) queue_work(ntos_wq, (work_struct))
-#define schedule_work(work_struct) queue_work(ntos_wq, (work_struct))
+#define schedule_ntos_work(work_struct) queue_work(ntos_wq, work_struct)
+#define schedule_work(work_struct) queue_work(ntos_wq, work_struct)
 #else
 #define schedule_ntos_work(work_struct) schedule_work(work_struct)
 #endif
 
-int ntoskernel_init(void);
-void ntoskernel_exit(void);
-int ntoskernel_init_device(struct wrap_device *wd);
-void ntoskernel_exit_device(struct wrap_device *wd);
-void *allocate_object(ULONG size, enum common_object_type type,
-		      struct unicode_string *name);
-void free_object(void *object);
+extern workqueue_struct_t *ndis_wq;
+#define schedule_ndis_work(work_struct) queue_work(ndis_wq, work_struct)
 
-int usb_init(void);
-void usb_exit(void);
-int usb_init_device(struct wrap_device *wd);
-void usb_exit_device(struct wrap_device *wd);
-void usb_cancel_pending_urbs(void);
+extern workqueue_struct_t *wrapndis_wq;
+#define schedule_wrapndis_work(work_struct) queue_work(wrapndis_wq, work_struct)
 
-int crt_init(void);
-void crt_exit(void);
-int rtl_init(void);
-void rtl_exit(void);
-int wrap_procfs_init(void);
-void wrap_procfs_remove(void);
+#define atomic_unary_op(var, size, oper)				\
+do {									\
+	if (size == 1)							\
+		__asm__ __volatile__(					\
+			LOCK_PREFIX oper "b %b0\n\t" : "+m" (var));	\
+	else if (size == 2)						\
+		__asm__ __volatile__(					\
+			LOCK_PREFIX oper "w %w0\n\t" : "+m" (var));	\
+	else if (size == 4)						\
+		__asm__ __volatile__(					\
+			LOCK_PREFIX oper "l %0\n\t" : "+m" (var));	\
+	else if (size == 8)						\
+		__asm__ __volatile__(					\
+			LOCK_PREFIX oper "q %q0\n\t" : "+m" (var));	\
+	else {								\
+		extern void _invalid_op_size_(void);			\
+		_invalid_op_size_();					\
+	}								\
+} while (0)
 
-int link_pe_images(struct pe_image *pe_image, unsigned short n);
+#define atomic_inc_var_size(var, size) atomic_unary_op(var, size, "inc")
 
-int stricmp(const char *s1, const char *s2);
-void dump_bytes(const char *name, const u8 *from, int len);
-struct mdl *allocate_init_mdl(void *virt, ULONG length);
-void free_mdl(struct mdl *mdl);
-struct driver_object *find_bus_driver(const char *name);
-void free_custom_extensions(struct driver_extension *drv_obj_ext);
-struct nt_thread *get_current_nt_thread(void);
-u64 ticks_1601(void);
-int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2);
-void wrap_init_timer(struct nt_timer *nt_timer, enum timer_type type,
-		     struct kdpc *kdpc, struct ndis_miniport_block *nmb);
-BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
-		       unsigned long repeat_hz, struct kdpc *kdpc);
+#define atomic_inc_var(var) atomic_inc_var_size(var, sizeof(var))
 
-LONG InterlockedDecrement(LONG volatile *val) wfastcall;
-LONG InterlockedIncrement(LONG volatile *val) wfastcall;
-struct nt_list *ExInterlockedInsertHeadList
-	(struct nt_list *head, struct nt_list *entry,
-	 NT_SPIN_LOCK *lock) wfastcall;
-struct nt_list *ExInterlockedInsertTailList
-	(struct nt_list *head, struct nt_list *entry,
-	 NT_SPIN_LOCK *lock) wfastcall;
-struct nt_list *ExInterlockedRemoveHeadList
-	(struct nt_list *head, NT_SPIN_LOCK *lock) wfastcall;
-NTSTATUS IofCallDriver(struct device_object *dev_obj, struct irp *irp) wfastcall;
-KIRQL KfRaiseIrql(KIRQL newirql) wfastcall;
-void KfLowerIrql(KIRQL oldirql) wfastcall;
-KIRQL KfAcquireSpinLock(NT_SPIN_LOCK *lock) wfastcall;
-void KfReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql) wfastcall;
-void IofCompleteRequest(struct irp *irp, CHAR prio_boost) wfastcall;
-void KefReleaseSpinLockFromDpcLevel(NT_SPIN_LOCK *lock) wfastcall;
+#define atomic_dec_var_size(var, size) atomic_unary_op(var, size, "dec")
 
-LONG ObfReferenceObject(void *object) wfastcall;
-void ObfDereferenceObject(void *object) wfastcall;
+#define atomic_dec_var(var) atomic_dec_var_size(var, sizeof(var))
 
-#define ObReferenceObject(object) ObfReferenceObject(object)
-#define ObDereferenceObject(object) ObfDereferenceObject(object)
+#define pre_atomic_add(var, i)					\
+({								\
+	typeof(var) pre;					\
+	__asm__ __volatile__(					\
+		LOCK_PREFIX "xadd %0, %1\n\t"			\
+		: "=r"(pre), "+m"(var)				\
+		: "0"(i));					\
+	pre;							\
+})
 
-void WRITE_PORT_UCHAR(ULONG_PTR port, UCHAR value) wstdcall;
-UCHAR READ_PORT_UCHAR(ULONG_PTR port) wstdcall;
+#define post_atomic_add(var, i) (pre_atomic_add(var, i) + i)
 
-#undef ExAllocatePoolWithTag
-void *ExAllocatePoolWithTag(enum pool_type pool_type, SIZE_T size,
-			    ULONG tag) wstdcall;
-#if defined(ALLOC_DEBUG) && ALLOC_DEBUG > 1
-#define ExAllocatePoolWithTag(pool_type, size, tag)			\
-	wrap_ExAllocatePoolWithTag(pool_type, size, tag, __FILE__, __LINE__)
+#ifndef in_atomic
+#define in_atomic() in_interrupt()
 #endif
 
-void ExFreePool(void *p) wstdcall;
-ULONG MmSizeOfMdl(void *base, ULONG length) wstdcall;
-void *MmMapIoSpace(PHYSICAL_ADDRESS phys_addr, SIZE_T size,
-		   enum memory_caching_type cache) wstdcall;
-void MmUnmapIoSpace(void *addr, SIZE_T size) wstdcall;
-void MmProbeAndLockPages(struct mdl *mdl, KPROCESSOR_MODE access_mode,
-			 enum lock_operation operation) wstdcall;
-void MmUnlockPages(struct mdl *mdl) wstdcall;
-void KeInitializeEvent(struct nt_event *nt_event,
-		       enum event_type type, BOOLEAN state) wstdcall;
-LONG KeSetEvent(struct nt_event *nt_event, KPRIORITY incr,
-		BOOLEAN wait) wstdcall;
-LONG KeResetEvent(struct nt_event *nt_event) wstdcall;
-void KeClearEvent(struct nt_event *nt_event) wstdcall;
-void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx) wstdcall;
-BOOLEAN queue_kdpc(struct kdpc *kdpc);
-BOOLEAN dequeue_kdpc(struct kdpc *kdpc);
+#ifndef preempt_enable_no_resched
+#define preempt_enable_no_resched() preempt_enable()
+#endif
 
-void KeFlushQueuedDpcs(void) wstdcall;
-NTSTATUS KeWaitForSingleObject(void *object, KWAIT_REASON reason,
-			       KPROCESSOR_MODE waitmode, BOOLEAN alertable,
-			       LARGE_INTEGER *timeout) wstdcall;
-struct mdl *IoAllocateMdl(void *virt, ULONG length, BOOLEAN second_buf,
-			  BOOLEAN charge_quota, struct irp *irp) wstdcall;
-void MmBuildMdlForNonPagedPool(struct mdl *mdl) wstdcall;
-void IoFreeMdl(struct mdl *mdl) wstdcall;
-NTSTATUS IoCreateDevice(struct driver_object *driver, ULONG dev_ext_length,
-			struct unicode_string *dev_name, DEVICE_TYPE dev_type,
-			ULONG dev_chars, BOOLEAN exclusive,
-			struct device_object **dev_obj) wstdcall;
-NTSTATUS IoCreateSymbolicLink(struct unicode_string *link,
-			      struct unicode_string *dev_name) wstdcall;
-void IoDeleteDevice(struct device_object *dev) wstdcall;
-void IoDetachDevice(struct device_object *topdev) wstdcall;
-struct device_object *IoGetAttachedDevice(struct device_object *dev) wstdcall;
-struct device_object *IoGetAttachedDeviceReference
-	(struct device_object *dev) wstdcall;
-NTSTATUS IoAllocateDriverObjectExtension
-	(struct driver_object *drv_obj, void *client_id, ULONG extlen,
-	 void **ext) wstdcall;
-void *IoGetDriverObjectExtension(struct driver_object *drv,
-				 void *client_id) wstdcall;
-struct device_object *IoAttachDeviceToDeviceStack
-	(struct device_object *src, struct device_object *dst) wstdcall;
-void KeInitializeEvent(struct nt_event *nt_event, enum event_type type,
-		       BOOLEAN state) wstdcall;
-struct irp *IoAllocateIrp(char stack_count, BOOLEAN charge_quota) wstdcall;
-void IoFreeIrp(struct irp *irp) wstdcall;
-BOOLEAN IoCancelIrp(struct irp *irp) wstdcall;
-struct irp *IoBuildSynchronousFsdRequest
-	(ULONG major_func, struct device_object *dev_obj, void *buf,
-	 ULONG length, LARGE_INTEGER *offset, struct nt_event *event,
-	 struct io_status_block *status) wstdcall;
-struct irp *IoBuildAsynchronousFsdRequest
-	(ULONG major_func, struct device_object *dev_obj, void *buf,
-	 ULONG length, LARGE_INTEGER *offset,
-	 struct io_status_block *status) wstdcall;
-NTSTATUS PoCallDriver(struct device_object *dev_obj, struct irp *irp) wstdcall;
+//#define DEBUG_IRQL 1
 
-NTSTATUS IoPassIrpDown(struct device_object *dev_obj, struct irp *irp) wstdcall;
-WIN_FUNC_DECL(IoPassIrpDown,2)
-NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
-			  struct irp *irp) wstdcall;
-NTSTATUS IoAsyncForwardIrp(struct device_object *dev_obj,
-			   struct irp *irp) wstdcall;
-NTSTATUS IoInvalidDeviceRequest(struct device_object *dev_obj,
-				struct irp *irp) wstdcall;
-
-KIRQL KeGetCurrentIrql(void) wstdcall;
-void KeInitializeSpinLock(NT_SPIN_LOCK *lock) wstdcall;
-void KeAcquireSpinLock(NT_SPIN_LOCK *lock, KIRQL *irql) wstdcall;
-void KeReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql) wstdcall;
-KIRQL KeAcquireSpinLockRaiseToDpc(NT_SPIN_LOCK *lock) wstdcall;
-
-void IoAcquireCancelSpinLock(KIRQL *irql) wstdcall;
-void IoReleaseCancelSpinLock(KIRQL irql) wstdcall;
-
-void RtlCopyMemory(void *dst, const void *src, SIZE_T length) wstdcall;
-NTSTATUS RtlUnicodeStringToAnsiString
-	(struct ansi_string *dst, const struct unicode_string *src,
-	 BOOLEAN dup) wstdcall;
-NTSTATUS RtlAnsiStringToUnicodeString
-	(struct unicode_string *dst, const struct ansi_string *src,
-	 BOOLEAN dup) wstdcall;
-void RtlInitAnsiString(struct ansi_string *dst, const char *src) wstdcall;
-void RtlInitString(struct ansi_string *dst, const char *src) wstdcall;
-void RtlInitUnicodeString(struct unicode_string *dest,
-			  const wchar_t *src) wstdcall;
-void RtlFreeUnicodeString(struct unicode_string *string) wstdcall;
-void RtlFreeAnsiString(struct ansi_string *string) wstdcall;
-LONG RtlCompareUnicodeString(const struct unicode_string *s1,
-			     const struct unicode_string *s2,
-			     BOOLEAN case_insensitive) wstdcall;
-void RtlCopyUnicodeString(struct unicode_string *dst,
-			  struct unicode_string *src) wstdcall;
-void KeInitializeTimer(struct nt_timer *nt_timer) wstdcall;
-void KeInitializeTimerEx(struct nt_timer *nt_timer,
-			 enum timer_type type) wstdcall;
-BOOLEAN KeSetTimerEx(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks,
-		     LONG period_ms, struct kdpc *kdpc) wstdcall;
-BOOLEAN KeSetTimer(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks,
-		   struct kdpc *kdpc) wstdcall;
-BOOLEAN KeCancelTimer(struct nt_timer *nt_timer) wstdcall;
-void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx) wstdcall;
-struct nt_thread *KeGetCurrentThread(void) wstdcall;
-NTSTATUS ObReferenceObjectByHandle(void *handle, ACCESS_MASK desired_access,
-				   void *obj_type, KPROCESSOR_MODE access_mode,
-				   void **object, void *handle_info) wstdcall;
-
-void adjust_user_shared_data_addr(char *driver, unsigned long length);
-
-#define IoCompleteRequest(irp, prio) IofCompleteRequest(irp, prio)
-#define IoCallDriver(dev, irp) IofCallDriver(dev, irp)
-
-#if defined(IO_DEBUG)
-#define DUMP_IRP(irp)							\
+#ifdef DEBUG_IRQL
+#define assert_irql(cond)						\
 do {									\
-	struct io_stack_location *irp_sl;				\
-	irp_sl = IoGetCurrentIrpStackLocation(irp);			\
-	IOTRACE("irp: %p, stack size: %d, cl: %d, sl: %p, dev_obj: %p, " \
-		"mj_fn: %d, minor_fn: %d, nt_urb: %p, event: %p",	\
-		irp, irp->stack_count, (irp)->current_location,		\
-		irp_sl, irp_sl->dev_obj, irp_sl->major_fn,		\
-		irp_sl->minor_fn, IRP_URB(irp),				\
-		(irp)->user_event);					\
+	KIRQL _irql_ = current_irql();					\
+	if (!(cond)) {							\
+		WARNING("assertion '%s' failed: %d", #cond, _irql_);	\
+		DBG_BLOCK(4) {						\
+			dump_stack();					\
+		}							\
+	}								\
 } while (0)
 #else
-#define DUMP_IRP(irp) do { } while (0)
+#define assert_irql(cond) do { } while (0)
 #endif
 
+/* When preempt is enabled, we should preempt_disable to raise IRQL to
+ * DISPATCH_LEVEL, to be consistent with the semantics. However, using
+ * a mutex instead, so that only ndiswrapper threads run one at a time
+ * on a processor when at DISPATCH_LEVEL seems to be enough. So that
+ * is what we will use until we learn otherwise. If
+ * preempt_(en|dis)able is required for some reason, comment out
+ * following #define. */
+
+#define WRAP_PREEMPT 1
+
+#if !defined(CONFIG_PREEMPT) || defined(CONFIG_PREEMPT_RT)
+#ifndef WRAP_PREEMPT
+#define WRAP_PREEMPT 1
+#endif
+#endif
+
+#ifdef WRAP_PREEMPT
+
+typedef struct {
+	int count;
+	struct mutex lock;
+#ifdef CONFIG_SMP
+	typeof(current->cpus_allowed) cpus_allowed;
+#endif
+	struct task_struct *task;
+} irql_info_t;
+
+DECLARE_PER_CPU(irql_info_t, irql_info);
+
+static inline KIRQL raise_irql(KIRQL newirql)
+{
+	irql_info_t *info;
+
+	assert(newirql == DISPATCH_LEVEL);
+	info = &get_cpu_var(irql_info);
+	if (info->task == current) {
+		assert(info->count > 0);
+		assert(mutex_is_locked(&info->lock));
+#if defined(CONFIG_SMP) && defined(DEBUG)
+		do {
+			cpumask_t cpumask;
+			cpumask = cpumask_of_cpu(smp_processor_id());
+			cpus_xor(cpumask, cpumask, current->cpus_allowed);
+			assert(cpus_empty(cpumask));
+		} while (0);
+#endif
+		info->count++;
+		put_cpu_var(irql_info);
+		return DISPATCH_LEVEL;
+	}
+	/* TODO: is this enough to pin down to current cpu? */
+#ifdef CONFIG_SMP
+	assert(task_cpu(current) == smp_processor_id());
+	info->cpus_allowed = current->cpus_allowed;
+	current->cpus_allowed = cpumask_of_cpu(smp_processor_id());
+#endif
+	put_cpu_var(irql_info);
+	mutex_lock(&info->lock);
+	assert(info->count == 0);
+	assert(info->task == NULL);
+	info->count = 1;
+	info->task = current;
+	return PASSIVE_LEVEL;
+}
+
+static inline void lower_irql(KIRQL oldirql)
+{									
+	irql_info_t *info;
+
+	assert(oldirql <= DISPATCH_LEVEL);
+	info = &get_cpu_var(irql_info);
+	assert(info->task == current);
+	assert(mutex_is_locked(&info->lock));
+	assert(info->count > 0);
+	if (--info->count == 0) {
+		info->task = NULL;
+#ifdef CONFIG_SMP
+		current->cpus_allowed = info->cpus_allowed;
+#endif
+		mutex_unlock(&info->lock);
+	}
+	put_cpu_var(irql_info);
+}
+
+static inline KIRQL current_irql(void)
+{
+	int count;
+	if (in_irq() || irqs_disabled())
+		EXIT4(return DIRQL);
+	if (in_atomic() || in_interrupt())
+		EXIT4(return SOFT_IRQL);
+	count = get_cpu_var(irql_info).count;
+	put_cpu_var(irql_info);
+	if (count)
+		EXIT6(return DISPATCH_LEVEL);
+	else
+		EXIT6(return PASSIVE_LEVEL);
+}
+
+#else
+
 static inline KIRQL current_irql(void)
 {
 	if (in_irq() || irqs_disabled())
-		EXIT6(return DIRQL);
+		EXIT4(return DIRQL);
 	if (in_interrupt())
-		EXIT6(return SOFT_IRQL);
+		EXIT4(return SOFT_IRQL);
 	if (in_atomic())
 		EXIT6(return DISPATCH_LEVEL);
 	else
@@ -822,27 +808,22 @@ static inline KIRQL current_irql(void)
 
 static inline KIRQL raise_irql(KIRQL newirql)
 {
-	KIRQL irql = current_irql();
-	TRACE6("%d, %d", irql, newirql);
-	assert(newirql <= DISPATCH_LEVEL);
-	assert(irql <= newirql);
-	if (irql < DISPATCH_LEVEL && newirql == DISPATCH_LEVEL)
-		preempt_disable();
-
-	return irql;
+	KIRQL ret = in_atomic() ? DISPATCH_LEVEL : PASSIVE_LEVEL;
+	assert(newirql == DISPATCH_LEVEL);
+	assert(current_irql() <= DISPATCH_LEVEL);
+	preempt_disable();
+	return ret;
 }
 
 static inline void lower_irql(KIRQL oldirql)
 {
-	KIRQL irql = current_irql();
-	TRACE6("%d, %d", irql, oldirql);
-	assert(irql <= DISPATCH_LEVEL);
-	assert(oldirql <= irql);
-	if (oldirql < DISPATCH_LEVEL && irql == DISPATCH_LEVEL)
-		preempt_enable();
+	assert(current_irql() == DISPATCH_LEVEL);
+	preempt_enable();
 }
 
-#define gfp_irql() (current_irql() < DISPATCH_LEVEL ? GFP_KERNEL : GFP_ATOMIC)
+#endif
+
+#define irql_gfp() (in_atomic() ? GFP_ATOMIC : GFP_KERNEL)
 
 /* Windows spinlocks are of type ULONG_PTR which is not big enough to
  * store Linux spinlocks; so we implement Windows spinlocks using
@@ -892,6 +873,10 @@ static inline void nt_spin_unlock(NT_SPIN_LOCK *lock)
 
 #endif // CONFIG_SMP
 
+/* When kernel would've disabled preempt (e.g., in interrupt
+ * handlers), we need to fake preempt so driver thinks it is running
+ * at right IRQL */
+
 /* raise IRQL to given (higher) IRQL if necessary before locking */
 static inline KIRQL nt_spin_lock_irql(NT_SPIN_LOCK *lock, KIRQL newirql)
 {
@@ -907,58 +892,20 @@ static inline void nt_spin_unlock_irql(NT_SPIN_LOCK *lock, KIRQL oldirql)
 	lower_irql(oldirql);
 }
 
-static inline void nt_spin_lock_bh(NT_SPIN_LOCK *lock)
-{
-	local_bh_disable();
-	preempt_disable();
-	nt_spin_lock(lock);
-}
-
-static inline void nt_spin_unlock_bh(NT_SPIN_LOCK *lock)
-{
-	nt_spin_unlock(lock);
-	preempt_enable_no_resched();
-	local_bh_enable();
-}
-
-#ifdef CONFIG_PREEMPT_RT
-#define save_local_irq(flags) raw_local_irq_save(flags)
-#define restore_local_irq(flags) raw_local_irq_restore(flags)
-#else
-#define save_local_irq(flags) local_irq_save(flags)
-#define restore_local_irq(flags) local_irq_restore(flags)
-#endif
-
 #define nt_spin_lock_irqsave(lock, flags)				\
 do {									\
-	save_local_irq(flags);						\
 	preempt_disable();						\
+	local_irq_save(flags);						\
 	nt_spin_lock(lock);						\
 } while (0)
 
 #define nt_spin_unlock_irqrestore(lock, flags)				\
 do {									\
 	nt_spin_unlock(lock);						\
-	restore_local_irq(flags);					\
+	local_irq_restore(flags);					\
 	preempt_enable();						\
 } while (0)
 
-#define atomic_insert_list_head(oldhead, head, newhead)			\
-	do {								\
-		oldhead = (typeof(oldhead))head;			\
-	} while (cmpxchg(&(head), oldhead, newhead) != (typeof(head))oldhead)
-
-#define atomic_remove_list_head(head, newhead)				\
-({									\
-	typeof(head) oldhead;						\
-	do {								\
-		oldhead = head;						\
-		if (!oldhead)						\
-			break;						\
-	} while (cmpxchg(&(head), oldhead, newhead) != oldhead);	\
-	oldhead;							\
-})
-
 static inline ULONG SPAN_PAGES(void *ptr, SIZE_T length)
 {
 	return PAGE_ALIGN(((unsigned long)ptr & (PAGE_SIZE - 1)) + length)
@@ -1057,4 +1004,221 @@ do {							\
 	schedule_timeout(n);				\
 } while (0)
 
+int ntoskernel_init(void);
+void ntoskernel_exit(void);
+int ntoskernel_init_device(struct wrap_device *wd);
+void ntoskernel_exit_device(struct wrap_device *wd);
+void *allocate_object(ULONG size, enum common_object_type type,
+		      struct unicode_string *name);
+void free_object(void *object);
+
+int usb_init(void);
+void usb_exit(void);
+int usb_init_device(struct wrap_device *wd);
+void usb_exit_device(struct wrap_device *wd);
+void usb_cancel_pending_urbs(void);
+
+int crt_init(void);
+void crt_exit(void);
+int rtl_init(void);
+void rtl_exit(void);
+int wrap_procfs_init(void);
+void wrap_procfs_remove(void);
+
+int link_pe_images(struct pe_image *pe_image, unsigned short n);
+
+int stricmp(const char *s1, const char *s2);
+void dump_bytes(const char *name, const u8 *from, int len);
+struct mdl *allocate_init_mdl(void *virt, ULONG length);
+void free_mdl(struct mdl *mdl);
+struct driver_object *find_bus_driver(const char *name);
+void free_custom_extensions(struct driver_extension *drv_obj_ext);
+struct nt_thread *get_current_nt_thread(void);
+u64 ticks_1601(void);
+int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2);
+void wrap_init_timer(struct nt_timer *nt_timer, enum timer_type type,
+		     struct ndis_mp_block *nmb);
+BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
+		       unsigned long repeat_hz, struct kdpc *kdpc);
+
+LONG InterlockedDecrement(LONG volatile *val) wfastcall;
+LONG InterlockedIncrement(LONG volatile *val) wfastcall;
+struct nt_list *ExInterlockedInsertHeadList
+	(struct nt_list *head, struct nt_list *entry,
+	 NT_SPIN_LOCK *lock) wfastcall;
+struct nt_list *ExInterlockedInsertTailList
+	(struct nt_list *head, struct nt_list *entry,
+	 NT_SPIN_LOCK *lock) wfastcall;
+struct nt_list *ExInterlockedRemoveHeadList
+	(struct nt_list *head, NT_SPIN_LOCK *lock) wfastcall;
+NTSTATUS IofCallDriver(struct device_object *dev_obj, struct irp *irp) wfastcall;
+KIRQL KfRaiseIrql(KIRQL newirql) wfastcall;
+void KfLowerIrql(KIRQL oldirql) wfastcall;
+KIRQL KfAcquireSpinLock(NT_SPIN_LOCK *lock) wfastcall;
+void KfReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql) wfastcall;
+void IofCompleteRequest(struct irp *irp, CHAR prio_boost) wfastcall;
+void KefReleaseSpinLockFromDpcLevel(NT_SPIN_LOCK *lock) wfastcall;
+
+LONG ObfReferenceObject(void *object) wfastcall;
+void ObfDereferenceObject(void *object) wfastcall;
+
+#define ObReferenceObject(object) ObfReferenceObject(object)
+#define ObDereferenceObject(object) ObfDereferenceObject(object)
+
+void WRITE_PORT_UCHAR(ULONG_PTR port, UCHAR value) wstdcall;
+UCHAR READ_PORT_UCHAR(ULONG_PTR port) wstdcall;
+
+#undef ExAllocatePoolWithTag
+void *ExAllocatePoolWithTag(enum pool_type pool_type, SIZE_T size,
+			    ULONG tag) wstdcall;
+#if defined(ALLOC_DEBUG) && ALLOC_DEBUG > 1
+#define ExAllocatePoolWithTag(pool_type, size, tag)			\
+	wrap_ExAllocatePoolWithTag(pool_type, size, tag, __FILE__, __LINE__)
+#endif
+
+void ExFreePool(void *p) wstdcall;
+ULONG MmSizeOfMdl(void *base, ULONG length) wstdcall;
+void *MmMapIoSpace(PHYSICAL_ADDRESS phys_addr, SIZE_T size,
+		   enum memory_caching_type cache) wstdcall;
+void MmUnmapIoSpace(void *addr, SIZE_T size) wstdcall;
+void MmProbeAndLockPages(struct mdl *mdl, KPROCESSOR_MODE access_mode,
+			 enum lock_operation operation) wstdcall;
+void MmUnlockPages(struct mdl *mdl) wstdcall;
+void KeInitializeEvent(struct nt_event *nt_event,
+		       enum event_type type, BOOLEAN state) wstdcall;
+LONG KeSetEvent(struct nt_event *nt_event, KPRIORITY incr,
+		BOOLEAN wait) wstdcall;
+LONG KeResetEvent(struct nt_event *nt_event) wstdcall;
+void KeClearEvent(struct nt_event *nt_event) wstdcall;
+void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx) wstdcall;
+BOOLEAN queue_kdpc(struct kdpc *kdpc);
+BOOLEAN dequeue_kdpc(struct kdpc *kdpc);
+
+void KeFlushQueuedDpcs(void) wstdcall;
+NTSTATUS IoConnectInterrupt(struct kinterrupt **kinterrupt,
+			    PKSERVICE_ROUTINE service_routine,
+			    void *service_context, NT_SPIN_LOCK *lock,
+			    ULONG vector, KIRQL irql, KIRQL synch_irql,
+			    enum kinterrupt_mode interrupt_mode,
+			    BOOLEAN shareable, KAFFINITY processor_enable_mask,
+			    BOOLEAN floating_save) wstdcall;
+void IoDisconnectInterrupt(struct kinterrupt *interrupt) wstdcall;
+BOOLEAN KeSynchronizeExecution(struct kinterrupt *interrupt,
+			       PKSYNCHRONIZE_ROUTINE synch_routine,
+			       void *ctx) wstdcall;
+
+NTSTATUS KeWaitForSingleObject(void *object, KWAIT_REASON reason,
+			       KPROCESSOR_MODE waitmode, BOOLEAN alertable,
+			       LARGE_INTEGER *timeout) wstdcall;
+struct mdl *IoAllocateMdl(void *virt, ULONG length, BOOLEAN second_buf,
+			  BOOLEAN charge_quota, struct irp *irp) wstdcall;
+void MmBuildMdlForNonPagedPool(struct mdl *mdl) wstdcall;
+void IoFreeMdl(struct mdl *mdl) wstdcall;
+NTSTATUS IoCreateDevice(struct driver_object *driver, ULONG dev_ext_length,
+			struct unicode_string *dev_name, DEVICE_TYPE dev_type,
+			ULONG dev_chars, BOOLEAN exclusive,
+			struct device_object **dev_obj) wstdcall;
+NTSTATUS IoCreateSymbolicLink(struct unicode_string *link,
+			      struct unicode_string *dev_name) wstdcall;
+void IoDeleteDevice(struct device_object *dev) wstdcall;
+void IoDetachDevice(struct device_object *topdev) wstdcall;
+struct device_object *IoGetAttachedDevice(struct device_object *dev) wstdcall;
+struct device_object *IoGetAttachedDeviceReference
+	(struct device_object *dev) wstdcall;
+NTSTATUS IoAllocateDriverObjectExtension
+	(struct driver_object *drv_obj, void *client_id, ULONG extlen,
+	 void **ext) wstdcall;
+void *IoGetDriverObjectExtension(struct driver_object *drv,
+				 void *client_id) wstdcall;
+struct device_object *IoAttachDeviceToDeviceStack
+	(struct device_object *src, struct device_object *dst) wstdcall;
+void KeInitializeEvent(struct nt_event *nt_event, enum event_type type,
+		       BOOLEAN state) wstdcall;
+struct irp *IoAllocateIrp(char stack_count, BOOLEAN charge_quota) wstdcall;
+void IoFreeIrp(struct irp *irp) wstdcall;
+BOOLEAN IoCancelIrp(struct irp *irp) wstdcall;
+struct irp *IoBuildSynchronousFsdRequest
+	(ULONG major_func, struct device_object *dev_obj, void *buf,
+	 ULONG length, LARGE_INTEGER *offset, struct nt_event *event,
+	 struct io_status_block *status) wstdcall;
+struct irp *IoBuildAsynchronousFsdRequest
+	(ULONG major_func, struct device_object *dev_obj, void *buf,
+	 ULONG length, LARGE_INTEGER *offset,
+	 struct io_status_block *status) wstdcall;
+NTSTATUS PoCallDriver(struct device_object *dev_obj, struct irp *irp) wstdcall;
+
+NTSTATUS IoPassIrpDown(struct device_object *dev_obj, struct irp *irp) wstdcall;
+WIN_FUNC_DECL(IoPassIrpDown,2);
+NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
+			  struct irp *irp) wstdcall;
+NTSTATUS IoAsyncForwardIrp(struct device_object *dev_obj,
+			   struct irp *irp) wstdcall;
+NTSTATUS IoInvalidDeviceRequest(struct device_object *dev_obj,
+				struct irp *irp) wstdcall;
+
+KIRQL KeGetCurrentIrql(void) wstdcall;
+void KeInitializeSpinLock(NT_SPIN_LOCK *lock) wstdcall;
+void KeAcquireSpinLock(NT_SPIN_LOCK *lock, KIRQL *irql) wstdcall;
+void KeReleaseSpinLock(NT_SPIN_LOCK *lock, KIRQL oldirql) wstdcall;
+KIRQL KeAcquireSpinLockRaiseToDpc(NT_SPIN_LOCK *lock) wstdcall;
+
+void IoAcquireCancelSpinLock(KIRQL *irql) wstdcall;
+void IoReleaseCancelSpinLock(KIRQL irql) wstdcall;
+
+void RtlCopyMemory(void *dst, const void *src, SIZE_T length) wstdcall;
+NTSTATUS RtlUnicodeStringToAnsiString
+	(struct ansi_string *dst, const struct unicode_string *src,
+	 BOOLEAN dup) wstdcall;
+NTSTATUS RtlAnsiStringToUnicodeString
+	(struct unicode_string *dst, const struct ansi_string *src,
+	 BOOLEAN dup) wstdcall;
+void RtlInitAnsiString(struct ansi_string *dst, const char *src) wstdcall;
+void RtlInitString(struct ansi_string *dst, const char *src) wstdcall;
+void RtlInitUnicodeString(struct unicode_string *dest,
+			  const wchar_t *src) wstdcall;
+void RtlFreeUnicodeString(struct unicode_string *string) wstdcall;
+void RtlFreeAnsiString(struct ansi_string *string) wstdcall;
+LONG RtlCompareUnicodeString(const struct unicode_string *s1,
+			     const struct unicode_string *s2,
+			     BOOLEAN case_insensitive) wstdcall;
+void RtlCopyUnicodeString(struct unicode_string *dst,
+			  struct unicode_string *src) wstdcall;
+NTSTATUS RtlUpcaseUnicodeString(struct unicode_string *dst,
+				struct unicode_string *src,
+				BOOLEAN alloc) wstdcall;
+void KeInitializeTimer(struct nt_timer *nt_timer) wstdcall;
+void KeInitializeTimerEx(struct nt_timer *nt_timer,
+			 enum timer_type type) wstdcall;
+BOOLEAN KeSetTimerEx(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks,
+		     LONG period_ms, struct kdpc *kdpc) wstdcall;
+BOOLEAN KeSetTimer(struct nt_timer *nt_timer, LARGE_INTEGER duetime_ticks,
+		   struct kdpc *kdpc) wstdcall;
+BOOLEAN KeCancelTimer(struct nt_timer *nt_timer) wstdcall;
+void KeInitializeDpc(struct kdpc *kdpc, void *func, void *ctx) wstdcall;
+struct nt_thread *KeGetCurrentThread(void) wstdcall;
+NTSTATUS ObReferenceObjectByHandle(void *handle, ACCESS_MASK desired_access,
+				   void *obj_type, KPROCESSOR_MODE access_mode,
+				   void **object, void *handle_info) wstdcall;
+
+void adjust_user_shared_data_addr(char *driver, unsigned long length);
+
+#define IoCompleteRequest(irp, prio) IofCompleteRequest(irp, prio)
+#define IoCallDriver(dev, irp) IofCallDriver(dev, irp)
+
+#if defined(IO_DEBUG)
+#define DUMP_IRP(irp)							\
+do {									\
+	struct io_stack_location *irp_sl;				\
+	irp_sl = IoGetCurrentIrpStackLocation(irp);			\
+	IOTRACE("irp: %p, stack size: %d, cl: %d, sl: %p, dev_obj: %p, " \
+		"mj_fn: %d, minor_fn: %d, nt_urb: %p, event: %p",	\
+		irp, irp->stack_count, (irp)->current_location,		\
+		irp_sl, irp_sl->dev_obj, irp_sl->major_fn,		\
+		irp_sl->minor_fn, IRP_URB(irp),				\
+		(irp)->user_event);					\
+} while (0)
+#else
+#define DUMP_IRP(irp) do { } while (0)
+#endif
+
 #endif // _NTOSKERNEL_H_
