@@ -19,31 +19,31 @@
 #include "usb.h"
 #include "loader.h"
 
-extern NT_SPIN_LOCK ntoskernel_lock;
-extern NT_SPIN_LOCK irp_cancel_lock;
+extern spinlock_t ntoskernel_lock;
+extern spinlock_t irp_cancel_lock;
 extern struct nt_list object_list;
 
 wstdcall void WIN_FUNC(IoAcquireCancelSpinLock,1)
 	(KIRQL *irql)
 {
-	*irql = nt_spin_lock_irql(&irp_cancel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&irp_cancel_lock);
+	*irql = 0;
 }
 
 wstdcall void WIN_FUNC(IoReleaseCancelSpinLock,1)
 	(KIRQL irql)
 {
-	nt_spin_unlock_irql(&irp_cancel_lock, irql);
+	spin_unlock_bh(&irp_cancel_lock);
 }
 
 wstdcall int WIN_FUNC(IoIsWdmVersionAvailable,2)
 	(UCHAR major, UCHAR minor)
 {
 	IOENTER("%d, %x", major, minor);
-	if ((major == 6 && minor == 0x00) || // Vista
-	    (major == 1 &&
-	     (minor == 0x30 || // Windows 2003
-	      minor == 0x20 || // Windows XP
-	      minor == 0x10))) // Windows 2000
+	if (major == 1 &&
+	    (minor == 0x30 || // Windows 2003
+	     minor == 0x20 || // Windows XP
+	     minor == 0x10)) // Windows 2000
 		IOEXIT(return TRUE);
 	IOEXIT(return FALSE);
 }
@@ -95,7 +95,7 @@ wstdcall struct irp *WIN_FUNC(IoAllocateIrp,2)
 	IOENTER("count: %d", stack_count);
 	stack_count++;
 	irp_size = IoSizeOfIrp(stack_count);
-	irp = kmalloc(irp_size, gfp_irql());
+	irp = kmalloc(irp_size, irql_gfp());
 	if (irp)
 		IoInitializeIrp(irp, irp_size, stack_count);
 	IOTRACE("irp %p", irp);
@@ -342,13 +342,13 @@ wfastcall NTSTATUS WIN_FUNC(IofCallDriver,2)
 	driver_dispatch_t *major_func;
 	struct driver_object *drv_obj;
 
-	IOTRACE("%p, %p, %p, %d, %d, %p", dev_obj, irp, dev_obj->drv_obj,
-		irp->current_location, irp->stack_count,
-		IoGetCurrentIrpStackLocation(irp));
 	if (irp->current_location <= 0) {
 		ERROR("invalid irp: %p, %d", irp, irp->current_location);
 		return STATUS_INVALID_PARAMETER;
 	}
+	IOTRACE("%p, %p, %p, %d, %d, %p", dev_obj, irp, dev_obj->drv_obj,
+		irp->current_location, irp->stack_count,
+		IoGetCurrentIrpStackLocation(irp));
 	IoSetNextIrpStackLocation(irp);
 	DUMP_IRP(irp);
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
@@ -369,6 +369,7 @@ wfastcall NTSTATUS WIN_FUNC(IofCallDriver,2)
 wfastcall void WIN_FUNC(IofCompleteRequest,2)
 	(struct irp *irp, CHAR prio_boost)
 {
+	struct io_stack_location *irp_sl;
 
 #ifdef IO_DEBUG
 	DUMP_IRP(irp);
@@ -376,17 +377,17 @@ wfastcall void WIN_FUNC(IofCompleteRequest,2)
 		ERROR("invalid irp: %p, STATUS_PENDING", irp);
 		return;
 	}
-	if (irp->current_location < 0) {
+	if (irp->current_location < 0 ||
+	    irp->current_location >= irp->stack_count) {
 		ERROR("invalid irp: %p, %d", irp, irp->current_location);
 		return;
 	}
 #endif
-	while (irp->current_location < irp->stack_count) {
-		struct io_stack_location *irp_sl;
+	for (irp_sl = IoGetCurrentIrpStackLocation(irp);
+	     irp->current_location < irp->stack_count; irp_sl++) {
 		struct device_object *dev_obj;
 		NTSTATUS status;
 
-		irp_sl = IoGetCurrentIrpStackLocation(irp);
 		DUMP_IRP(irp);
 		if (irp_sl->control & SL_PENDING_RETURNED)
 			irp->pending_returned = TRUE;
@@ -397,7 +398,6 @@ wfastcall void WIN_FUNC(IofCompleteRequest,2)
 		 * is what we are going to call below; so we set
 		 * current_location and dev_obj for the previous
 		 * (higher) location */
-
 		IoSkipCurrentIrpStackLocation(irp);
 		if (irp->current_location < irp->stack_count)
 			dev_obj = IoGetCurrentIrpStackLocation(irp)->dev_obj;
@@ -524,46 +524,44 @@ static irqreturn_t io_irq_isr(int irq, void *data ISR_PT_REGS_PARAM_DECL)
 	struct kinterrupt *interrupt = data;
 	BOOLEAN ret;
 
-	nt_spin_lock(&interrupt->lock);
-	ret = LIN2WIN2(interrupt->service_routine, interrupt,
-		       interrupt->service_context);
-	nt_spin_unlock(&interrupt->lock);
+	TRACE6("%p", interrupt);
+	nt_spin_lock(interrupt->actual_lock);
+	ret = LIN2WIN2(interrupt->isr, interrupt, interrupt->isr_ctx);
+	nt_spin_unlock(interrupt->actual_lock);
 	if (ret == TRUE)
-		return IRQ_HANDLED;
+		EXIT6(return IRQ_HANDLED);
 	else
-		return IRQ_NONE;
+		EXIT6(return IRQ_NONE);
 }
 
 wstdcall NTSTATUS WIN_FUNC(IoConnectInterrupt,11)
-	(struct kinterrupt **kinterrupt, PKSERVICE_ROUTINE service_routine,
-	 void *service_context, NT_SPIN_LOCK *lock, ULONG vector,
-	 KIRQL irql, KIRQL synch_irql, enum kinterrupt_mode interrupt_mode,
-	 BOOLEAN shareable, KAFFINITY processor_enable_mask,
-	 BOOLEAN floating_save)
+	(struct kinterrupt **kinterrupt, PKSERVICE_ROUTINE isr, void *isr_ctx,
+	 NT_SPIN_LOCK *lock, ULONG vector, KIRQL irql, KIRQL synch_irql,
+	 enum kinterrupt_mode mode, BOOLEAN shared, KAFFINITY cpu_mask,
+	 BOOLEAN save_fp)
 {
 	struct kinterrupt *interrupt;
 	IOENTER("");
-	interrupt = kmalloc(sizeof(*interrupt), GFP_KERNEL);
+	interrupt = kzalloc(sizeof(*interrupt), GFP_KERNEL);
 	if (!interrupt)
 		IOEXIT(return STATUS_INSUFFICIENT_RESOURCES);
-	memset(interrupt, 0, sizeof(*interrupt));
 	interrupt->vector = vector;
-	interrupt->processor_enable_mask = processor_enable_mask;
+	interrupt->cpu_mask = cpu_mask;
 	nt_spin_lock_init(&interrupt->lock);
 	if (lock)
 		interrupt->actual_lock = lock;
 	else
 		interrupt->actual_lock = &interrupt->lock;
-	interrupt->shareable = shareable;
-	interrupt->floating_save = floating_save;
-	interrupt->service_routine = service_routine;
-	interrupt->service_context = service_context;
+	interrupt->shared = shared;
+	interrupt->save_fp = save_fp;
+	interrupt->isr = isr;
+	interrupt->isr_ctx = isr_ctx;
 	InitializeListHead(&interrupt->list);
 	interrupt->irql = irql;
 	interrupt->synch_irql = synch_irql;
-	interrupt->interrupt_mode = interrupt_mode;
-	if (request_irq(vector, io_irq_isr, shareable ? IRQF_SHARED : 0,
-			"io_irq", interrupt)) {
+	interrupt->mode = mode;
+	if (request_irq(vector, io_irq_isr, shared ? IRQF_SHARED : 0,
+			"ndiswrapper", interrupt)) {
 		WARNING("request for irq %d failed", vector);
 		kfree(interrupt);
 		IOEXIT(return STATUS_INSUFFICIENT_RESOURCES);
@@ -615,7 +613,7 @@ wstdcall struct io_workitem *WIN_FUNC(IoAllocateWorkItem,1)
 	struct io_workitem *io_workitem;
 
 	IOENTER("%p", dev_obj);
-	io_workitem = kmalloc(sizeof(*io_workitem), gfp_irql());
+	io_workitem = kmalloc(sizeof(*io_workitem), irql_gfp());
 	if (!io_workitem)
 		IOEXIT(return NULL);
 	io_workitem->dev_obj = dev_obj;
@@ -653,18 +651,17 @@ wstdcall NTSTATUS WIN_FUNC(IoAllocateDriverObjectExtension,4)
 	 void **ext)
 {
 	struct custom_ext *ce;
-	KIRQL irql;
 
 	IOENTER("%p, %p", drv_obj, client_id);
-	ce = kmalloc(sizeof(*ce) + extlen, gfp_irql());
+	ce = kmalloc(sizeof(*ce) + extlen, irql_gfp());
 	if (ce == NULL)
 		return STATUS_INSUFFICIENT_RESOURCES;
 
 	IOTRACE("custom_ext: %p", ce);
 	ce->client_id = client_id;
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&ntoskernel_lock);
 	InsertTailList(&drv_obj->drv_ext->custom_ext, &ce->list);
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	spin_unlock_bh(&ntoskernel_lock);
 
 	*ext = (void *)ce + sizeof(*ce);
 	IOTRACE("ext: %p", *ext);
@@ -676,18 +673,17 @@ wstdcall void *WIN_FUNC(IoGetDriverObjectExtension,2)
 {
 	struct custom_ext *ce;
 	void *ret;
-	KIRQL irql;
 
 	IOENTER("drv_obj: %p, client_id: %p", drv_obj, client_id);
 	ret = NULL;
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&ntoskernel_lock);
 	nt_list_for_each_entry(ce, &drv_obj->drv_ext->custom_ext, list) {
 		if (ce->client_id == client_id) {
 			ret = (void *)ce + sizeof(*ce);
 			break;
 		}
 	}
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	spin_unlock_bh(&ntoskernel_lock);
 	IOTRACE("ret: %p", ret);
 	return ret;
 }
@@ -695,13 +691,12 @@ wstdcall void *WIN_FUNC(IoGetDriverObjectExtension,2)
 void free_custom_extensions(struct driver_extension *drv_ext)
 {
 	struct nt_list *ent;
-	KIRQL irql;
 
 	IOENTER("%p", drv_ext);
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&ntoskernel_lock);
 	while ((ent = RemoveHeadList(&drv_ext->custom_ext)))
 		kfree(ent);
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	spin_unlock_bh(&ntoskernel_lock);
 	IOEXIT(return);
 }
 
@@ -821,7 +816,6 @@ wstdcall void WIN_FUNC(IoDetachDevice,1)
 	(struct device_object *tgt)
 {
 	struct device_object *tail;
-	KIRQL irql;
 
 	IOENTER("%p", tgt);
 	if (!tgt)
@@ -831,29 +825,27 @@ wstdcall void WIN_FUNC(IoDetachDevice,1)
 		IOEXIT(return);
 	IOTRACE("tail: %p", tail);
 
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&ntoskernel_lock);
 	tgt->attached = tail->attached;
 	IOTRACE("attached:%p", tgt->attached);
 	for ( ; tail; tail = tail->attached) {
 		IOTRACE("tail:%p", tail);
 		tail->stack_count--;
 	}
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	spin_unlock_bh(&ntoskernel_lock);
 	IOEXIT(return);
 }
 
 wstdcall struct device_object *WIN_FUNC(IoGetAttachedDevice,1)
 	(struct device_object *dev)
 {
-	KIRQL irql;
-
 	IOENTER("%p", dev);
 	if (!dev)
 		IOEXIT(return NULL);
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&ntoskernel_lock);
 	while (dev->attached)
 		dev = dev->attached;
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	spin_unlock_bh(&ntoskernel_lock);
 	IOEXIT(return dev);
 }
 
@@ -874,19 +866,17 @@ wstdcall struct device_object *WIN_FUNC(IoAttachDeviceToDeviceStack,2)
 	struct device_object *attached;
 	struct dev_obj_ext *src_dev_ext;
 
-	KIRQL irql;
-
 	IOENTER("%p, %p", src, tgt);
 	attached = IoGetAttachedDevice(tgt);
 	IOTRACE("%p", attached);
 	src_dev_ext = src->dev_obj_ext;
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&ntoskernel_lock);
 	if (attached)
 		attached->attached = src;
 	src->attached = NULL;
 	src->stack_count = attached->stack_count + 1;
 	src_dev_ext->attached_to = attached;
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	spin_unlock_bh(&ntoskernel_lock);
 	IOTRACE("stack_count: %d -> %d", attached->stack_count,
 		src->stack_count);
 	IOEXIT(return attached);
@@ -937,13 +927,12 @@ wstdcall NTSTATUS WIN_FUNC(IoGetDeviceObjectPointer,4)
 	(struct unicode_string *name, ACCESS_MASK desired_access,
 	 void *file_obj, struct device_object *dev_obj)
 {
-	KIRQL irql;
 	struct common_object_header *coh;
 
 	dev_obj = NULL;
 	/* TODO: access is not checked and file_obj is set to NULL */
 	file_obj = NULL;
-	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
+	spin_lock_bh(&ntoskernel_lock);
 	nt_list_for_each_entry(coh, &object_list, list) {
 		TRACE5("header: %p, type: %d", coh, coh->type);
 		if (coh->type != OBJECT_TYPE_DEVICE)
@@ -954,7 +943,7 @@ wstdcall NTSTATUS WIN_FUNC(IoGetDeviceObjectPointer,4)
 			break;
 		}
 	}
-	nt_spin_unlock_irql(&ntoskernel_lock, irql);
+	spin_unlock_bh(&ntoskernel_lock);
 	if (dev_obj)
 		IOEXIT(return STATUS_SUCCESS);
 	else
@@ -1021,7 +1010,7 @@ wstdcall void *WIN_FUNC(IoAllocateErrorLogEntry,2)
 {
 	/* not implemented fully */
 	void *ret = kmalloc(sizeof(struct io_error_log_packet) + entry_size,
-			    gfp_irql());
+			    irql_gfp());
 	TRACE2("%p", ret);
 	if (ret)
 		return ret + sizeof(struct io_error_log_packet);
