@@ -20,12 +20,6 @@
 #include "ndis.h"
 #include "wrapndis.h"
 
-driver_dispatch_t IopInvalidDeviceRequest;
-driver_dispatch_t IopPassIrpDown;
-
-STDCALL NTSTATUS IrpStopCompletion(struct device_object *dev_obj,
-				   struct irp *irp, void *context);
-
 NTSTATUS pnp_set_power_state(struct wrap_device *wd,
 			     enum device_power_state state);
 NTSTATUS pnp_start_device(struct wrap_device *wd);
