@@ -233,7 +233,7 @@ struct nt_list {
 	struct nt_list *prev;
 };
 
-typedef volatile ULONG_PTR NT_SPIN_LOCK;
+typedef ULONG_PTR NT_SPIN_LOCK;
 
 enum kdpc_importance {LowImportance, MediumImportance, HighImportance};
 
@@ -1080,9 +1080,9 @@ struct irp {
 };
 
 #define IoSizeOfIrp(stack_count)					\
-	((USHORT)(sizeof(struct irp) + ((stack_count) *			\
-					sizeof(struct io_stack_location))))
-#define IoGetCurrentIrpStackLocation(irp)		\
+	((USHORT)(sizeof(struct irp) +					\
+		  ((stack_count) * sizeof(struct io_stack_location))))
+#define IoGetCurrentIrpStackLocation(irp)	\
 	(irp)->tail.overlay.csl
 #define IoGetNextIrpStackLocation(irp)		\
 	(IoGetCurrentIrpStackLocation(irp) - 1)
@@ -1318,18 +1318,18 @@ typedef BOOLEAN (*PKSYNCHRONIZE_ROUTINE)(void *context) wstdcall;
 
 struct kinterrupt {
 	ULONG vector;
-	KAFFINITY processor_enable_mask;
+	KAFFINITY cpu_mask;
 	NT_SPIN_LOCK lock;
 	NT_SPIN_LOCK *actual_lock;
-	BOOLEAN shareable;
-	BOOLEAN floating_save;
+	BOOLEAN shared;
+	BOOLEAN save_fp;
 	CHAR processor_number;
-	PKSERVICE_ROUTINE service_routine;
-	void *service_context;
+	PKSERVICE_ROUTINE isr;
+	void *isr_ctx;
 	struct nt_list list;
 	KIRQL irql;
 	KIRQL synch_irql;
-	enum kinterrupt_mode interrupt_mode;
+	enum kinterrupt_mode mode;
 };
 
 struct time_fields {
