@@ -21,6 +21,7 @@
 char *if_name = "wlan%d";
 int proc_uid, proc_gid;
 int hangcheck_interval;
+char *utils_version = UTILS_VERSION;
 
 #if defined(DEBUG) && (DEBUG > 0)
 int debug = DEBUG;
@@ -48,6 +49,10 @@ WRAP_MODULE_PARM_INT(hangcheck_interval, 0600);
 MODULE_PARM_DESC(hangcheck_interval, "The interval, in seconds, for checking"
 		 " if driver is hung. (default: 0)");
 
+WRAP_MODULE_PARM_STRING(utils_version, 0400);
+MODULE_PARM_DESC(utils_version, "Compatible version of utils "
+		 "(read only: " UTILS_VERSION ")");
+
 MODULE_AUTHOR("ndiswrapper team <ndiswrapper-general@lists.sourceforge.net>");
 #ifdef MODULE_DESCRIPTION
 MODULE_DESCRIPTION("NDIS wrapper driver");
@@ -75,12 +80,20 @@ static void module_cleanup(void)
 
 static int __init wrapper_init(void)
 {
-	printk(KERN_INFO "%s version %s loaded (smp=%s)\n",
+	printk(KERN_INFO "%s version %s loaded (smp=%s, preempt=%s)\n",
 	       DRIVER_NAME, DRIVER_VERSION,
 #ifdef CONFIG_SMP
 	       "yes"
 #else
 	       "no"
+#endif
+		,
+#ifdef CONFIG_PREEMPT_RT
+		"rt"
+#elif defined(CONFIG_PREEMPT)
+		"yes"
+#else
+		"no"
 #endif
 		);
 
